from doitpy.pyflakes import Pyflakes
from doitpy.coverage import Coverage, PythonPackage


DOIT_CONFIG = {
    'default_tasks': ['pyflakes'],
    }


def task_pyflakes():
    flaker = Pyflakes()
    yield flaker.tasks('*.py')
    yield flaker.tasks('import_deps/*.py')
    yield flaker.tasks('tests/*.py')


def task_coverage():
    """show coverage for all modules including tests"""
    cov = Coverage(
        [PythonPackage('import_deps', 'tests')],
        config={'branch':True,},
    )
    yield cov.all() # create task `coverage`
    yield cov.src() # create task `coverage_src`



__version__ = (0, 4, 0)

import ast
import pathlib


class _ImportsFinder(ast.NodeVisitor):
    """find all imports
    :ivar imports: (list - tuple) (module, name, asname, level)
    """
    def __init__(self):
        ast.NodeVisitor.__init__(self)
        self.imports = []

    def visit_Import(self, node):
        """callback for 'import' statement"""
        self.imports.extend((None, n.name, n.asname, None)
                            for n in node.names)
        ast.NodeVisitor.generic_visit(self, node)

    def visit_ImportFrom(self, node):
        """callback for 'import from' statement"""
        self.imports.extend((node.module, n.name, n.asname, node.level)
                            for n in node.names)
        ast.NodeVisitor.generic_visit(self, node)

def ast_imports(file_path):
    """get list of import from python module
    :return: (list - tuple) (module, name, asname, level)
    """
    with pathlib.Path(file_path).open('r') as fp:
        text = fp.read()
    mod_ast = ast.parse(text, str(file_path))
    finder = _ImportsFinder()
    finder.visit(mod_ast)
    return finder.imports


##########


class PyModule(object):
    """Represents a python module

    :ivar path: (pathlib.Path) module's path
    :ivar fqn: (list - str) full qualified name as list of strings
    """
    def __init__(self, path):
        self.path = pathlib.Path(path)
        assert self.path.suffix == '.py'
        self.fqn = self._get_fqn(self.path)

    def __repr__(self):
        return "<PyModule {}>".format(self.path)

    @staticmethod
    def is_pkg(path):
        """return True if path is a python package"""
        return (path.is_dir() and (path / '__init__.py').exists())

    def pkg_path(self):
        """return pathlib.Path that contains top-most package/module
        Path that is supposed to be part of PYTHONPATH
        """
        return self.path.parents[len(self.fqn)-1]

    @classmethod
    def _get_fqn(cls, path):
        """get full qualified name as list of strings
        :return: (list - str) of path segments from top package to given path
        """
        name_list = [path.stem]
        current_path = path
        # move to parent path until parent path is a python package
        while True:
            parent = current_path.parent
            if parent.name in ('', '.', '..'):
                break
            if not cls.is_pkg(parent):
                break
            name_list.append(parent.name)
            current_path = parent
        return list(reversed(name_list))



class ModuleSet(object):
    """helper to filter import list only from within packages"""
    def __init__(self, path_list):
        self.pkgs = set() # str of fqn (dot separed)
        self.by_path = {} # module by path
        self.by_name = {} # module by name (dot separated)

        for path in path_list:
            # create modules object
            mod = PyModule(path)
            if mod.fqn[-1] == '__init__':
                self.pkgs.add('.'.join(mod.fqn[:-1]))
            self.by_path[path] = mod
            self.by_name['.'.join(mod.fqn)] = mod


    def _get_imported_module(self, module_name):
        """try to get imported module reference by its name"""
        # if imported module on module_set add to list
        imp_mod = self.by_name.get(module_name)
        if imp_mod:
            return imp_mod

        # last part of import section might not be a module
        # remove last section
        no_obj = module_name.rsplit('.', 1)[0]
        imp_mod2 = self.by_name.get(no_obj)
        if imp_mod2:
            return imp_mod2

        # special case for __init__
        if module_name in self.pkgs:
            pkg_name = module_name  + ".__init__"
            return self.by_name[pkg_name]

        if no_obj in self.pkgs:
            pkg_name = no_obj +  ".__init__"
            return self.by_name[pkg_name]


    def get_imports(self, module, return_fqn=False):
        """return set of imported modules that are in self
        :param module: PyModule
        :return: (set - Path)
                 (set - str) if return_fqn == True
        """
        # print('####', module.fqn)
        # print(self.by_name.keys(), '\n\n')
        imports = set()
        raw_imports = ast_imports(module.path)
        for import_entry in raw_imports:
            # join 'from' and 'import' part of import statement
            full = ".".join(s for s in import_entry[:2] if s)

            import_level = import_entry[3]
            if import_level:
                # intra package imports
                intra = '.'.join(module.fqn[:-import_level] + [full])
                imported = self._get_imported_module(intra)
            else:
                imported = self._get_imported_module(full)

            if imported:
                if return_fqn:
                    imports.add('.'.join(imported.fqn))
                else:
                    imports.add(imported.path)
        return imports


    # higher level API
    def mod_imports(self, mod_fqn):
        mod = self.by_name[mod_fqn]
        return self.get_imports(mod, return_fqn=True)



import argparse
import json
import pathlib
import sys

from . import __version__, PyModule, ModuleSet


def detect_cycles(results):
    """Detect circular dependencies using DFS
    Returns set of edges (module, import) that create cycles
    """
    # Build adjacency list
    graph = {}
    for result in results:
        module = result['module']
        graph[module] = result['imports']

    cycle_edges = set()
    visited = set()
    rec_stack = set()

    def dfs(node, path):
        visited.add(node)
        rec_stack.add(node)
        path.append(node)

        for neighbor in graph.get(node, []):
            if neighbor not in visited:
                if neighbor in graph:  # Only follow if it's in our tracked modules
                    dfs(neighbor, path)
            elif neighbor in rec_stack:
                # Found a cycle - mark all edges in the cycle
                cycle_start_idx = path.index(neighbor)
                for i in range(cycle_start_idx, len(path)):
                    if i + 1 < len(path):
                        cycle_edges.add((path[i], path[i + 1]))
                # Add the back edge
                cycle_edges.add((node, neighbor))

        rec_stack.remove(node)
        path.pop()

    for module in graph:
        if module not in visited:
            dfs(module, [])

    return cycle_edges


def topological_sort(results):
    """Topological sort of modules (dependencies before dependents).
    Uses Kahn's algorithm with rank-based ordering for stability.
    Rank is defined as the longest path from any leaf node (module that imports but isn't imported).
    When multiple nodes become available, nodes with higher rank are output first.

    Handles circular dependencies gracefully:
    - Nodes actually in cycles are identified and processed last
    - Nodes that depend on cycles (but aren't in them) are processed normally
    - Isolated nodes (no dependencies, no dependents) are placed last

    Returns list of module names in topological order (dependencies before dependents).
    """
    # Collect all modules
    all_modules = set(result['module'] for result in results)

    # Build dependencies: module -> list of modules it imports (its dependencies)
    dependencies = {module: [] for module in all_modules}
    for result in results:
        module = result['module']
        dependencies[module] = [imp for imp in result['imports'] if imp in all_modules]

    # Build reverse graph: module -> list of modules that import it (its dependents)
    dependents = {module: [] for module in all_modules}

    for module in all_modules:
        for dep in dependencies[module]:
            dependents[dep].append(module)

    # Calculate rank for each node (longest path from any leaf node)
    # Leaf nodes are those that have no dependents (nothing imports them)
    # Detect cycles using DFS with recursion stack
    rank = {}
    in_cycle = set()

    def calculate_rank(node, visiting=None, rec_path=None):
        if visiting is None:
            visiting = set()
        if rec_path is None:
            rec_path = []

        if node in rank:
            return rank[node]

        if node in visiting:
            # Cycle detected - mark all nodes in the cycle path
            cycle_start = rec_path.index(node)
            for i in range(cycle_start, len(rec_path)):
                in_cycle.add(rec_path[i])
            in_cycle.add(node)
            return -1  # Special value for cycles

        visiting.add(node)
        rec_path.append(node)
        deps = dependents[node]  # Use dependents (who imports this node)

        if not deps:
            rank[node] = 1  # Leaf nodes (not imported by anyone) have rank 1
        else:
            dep_ranks = []
            for dep in deps:
                dep_rank = calculate_rank(dep, visiting, rec_path)
                if dep_rank == -1:
                    # Dependent is in cycle
                    pass
                else:
                    dep_ranks.append(dep_rank)

            # Only mark as cycle if this node is actually in the cycle
            if node in in_cycle:
                rank[node] = -1
            elif dep_ranks:
                rank[node] = max(dep_ranks) + 1
            else:
                # All dependents are in cycles, but this node isn't
                rank[node] = 2

        rec_path.pop()
        visiting.remove(node)
        return rank[node]

    for module in all_modules:
        if module not in rank:
            calculate_rank(module)

    # Topological sort: start with roots (nodes with no dependencies)
    # in_degree tracks how many unprocessed dependencies each node has
    in_degree = {module: len(dependencies[module]) for module in all_modules}

    # Separate cycle nodes and isolated nodes from regular nodes
    cycle_nodes = {node for node in all_modules if rank[node] == -1}
    isolated_nodes = {node for node in all_modules
                      if len(dependencies[node]) == 0 and len(dependents[node]) == 0}

    non_cycle_roots = [node for node in all_modules
                       if in_degree[node] == 0
                       and node not in cycle_nodes
                       and node not in isolated_nodes]

    # Initial queue: non-cycle, non-isolated roots, sorted by rank DESC then name ASC
    queue = sorted(non_cycle_roots, key=lambda x: (-rank[x], x))
    sorted_list = []

    while queue:
        node = queue.pop(0)
        sorted_list.append(node)

        # Process all dependents of this node (nodes that import this node)
        for dependent in dependents[node]:
            if dependent not in cycle_nodes and dependent not in isolated_nodes:
                in_degree[dependent] -= 1
                if in_degree[dependent] == 0:
                    # Insert maintaining rank order (higher rank first)
                    # Within same rank, maintain FIFO order
                    dep_rank = rank[dependent]
                    insert_idx = len(queue)
                    for i, queued_node in enumerate(queue):
                        if rank[queued_node] < dep_rank:
                            insert_idx = i
                            break
                    queue.insert(insert_idx, dependent)

    # Handle remaining nodes (cycles and nodes not yet processed, but not isolated)
    remaining = all_modules - set(sorted_list) - isolated_nodes
    if remaining:
        # Add remaining nodes sorted alphabetically
        sorted_list.extend(sorted(remaining))

    # Add isolated nodes last (sorted alphabetically)
    if isolated_nodes:
        sorted_list.extend(sorted(isolated_nodes))

    return sorted_list


def format_dot(results, highlight_cycles=True):
    """Format results as DOT graph for graphviz"""
    lines = ['digraph imports {']
    lines.append('    rankdir=LR;')
    lines.append('    node [shape=box, style="rounded,filled", fillcolor=lightblue, fontname="Arial"];')
    lines.append('    edge [fontname="Arial"];')

    # Detect cycles
    cycle_edges = detect_cycles(results) if highlight_cycles else set()

    # Group modules by package
    packages = {}
    all_modules = set()

    for result in results:
        module = result['module']
        all_modules.add(module)
        # Extract package hierarchy
        parts = module.split('.')
        if len(parts) > 1:
            # Get package path (everything except last part)
            pkg = '.'.join(parts[:-1])
            if pkg not in packages:
                packages[pkg] = []
            packages[pkg].append(module)

    # Create subgraphs for packages
    def create_subgraph(pkg_name, modules, indent=1):
        ind = '    ' * indent
        lines.append(f'{ind}subgraph cluster_{pkg_name.replace(".", "_")} {{')
        lines.append(f'{ind}    label = "{pkg_name}";')
        lines.append(f'{ind}    style = "rounded,dashed";')
        lines.append(f'{ind}    color = gray40;')
        lines.append(f'{ind}    fontsize = 11;')
        lines.append(f'{ind}    fontcolor = gray20;')
        lines.append(f'{ind}    penwidth = 1.5;')

        # Find direct children of this package
        for mod in sorted(modules):
            if mod.rsplit('.', 1)[0] == pkg_name:
                lines.append(f'{ind}    "{mod}";')

        # Find sub-packages
        sub_pkgs = {}
        for other_pkg, other_modules in packages.items():
            if other_pkg.startswith(pkg_name + '.') and other_pkg.count('.') == pkg_name.count('.') + 1:
                sub_pkgs[other_pkg] = other_modules

        for sub_pkg in sorted(sub_pkgs.keys()):
            create_subgraph(sub_pkg, sub_pkgs[sub_pkg], indent + 1)

        lines.append(f'{ind}}}')

    # Create top-level packages
    top_level_pkgs = set()
    for pkg in packages:
        top = pkg.split('.')[0]
        top_level_pkgs.add(top)

    for top_pkg in sorted(top_level_pkgs):
        pkg_modules = [m for pkg, modules in packages.items()
                       if pkg.startswith(top_pkg)
                       for m in modules]
        if pkg_modules:
            create_subgraph(top_pkg, pkg_modules)

    # Add edges with cycle detection
    lines.append('')
    for result in results:
        module = result['module']

        for imp in result['imports']:
            # Check if this edge is part of a cycle
            if (module, imp) in cycle_edges:
                lines.append(f'    "{module}" -> "{imp}" [color=red, penwidth=2.0];')
            else:
                lines.append(f'    "{module}" -> "{imp}";')

    lines.append('}')
    return '\n'.join(lines)


def main(argv=sys.argv):
    parser = argparse.ArgumentParser(prog='import_deps')
    parser.add_argument('path', metavar='PATH',
                        help='Python file or package directory to analyze')
    parser.add_argument('--json', action='store_true',
                        help='Output results in JSON format')
    parser.add_argument('--dot', action='store_true',
                        help='Output results in DOT format for graphviz')
    parser.add_argument('--check', action='store_true',
                        help='Check for circular dependencies and exit with error if found')
    parser.add_argument('--sort', action='store_true',
                        help='Output modules in topological sort order (dependencies first)')
    parser.add_argument('--version', action='version',
                        version='.'.join(str(i) for i in __version__))
    config = parser.parse_args(argv[1:])

    # Check for mutually exclusive flags
    output_flags = sum([config.json, config.dot, config.sort])
    if output_flags > 1:
        print("Error: --json, --dot, and --sort are mutually exclusive", file=sys.stderr)
        sys.exit(1)

    path = pathlib.Path(config.path)

    # Collect data
    if path.is_file():
        # Single file analysis
        module = PyModule(config.path)
        base_path = module.pkg_path().resolve()
        mset = ModuleSet(base_path.glob('**/*.py'))
        imports = mset.get_imports(module, return_fqn=True)

        results = [{
            'module': '.'.join(module.fqn),
            'imports': sorted(imports)
        }]

    elif path.is_dir():
        # Package analysis
        base_path = path.resolve()
        py_files = list(base_path.glob('**/*.py'))
        mset = ModuleSet(py_files)

        results = []
        for mod_name in sorted(mset.by_name.keys()):
            mod = mset.by_name[mod_name]
            imports = mset.get_imports(mod, return_fqn=True)
            results.append({
                'module': mod_name,
                'imports': sorted(imports)
            })

    else:
        print(f"Error: {config.path} is not a valid file or directory", file=sys.stderr)
        sys.exit(1)

    # Check for circular dependencies
    if config.check:
        cycle_edges = detect_cycles(results)
        if cycle_edges:
            print("Circular dependencies detected:", file=sys.stderr)

            # Group cycles by modules involved
            cycles_by_module = {}
            for src, dst in cycle_edges:
                if src not in cycles_by_module:
                    cycles_by_module[src] = []
                cycles_by_module[src].append(dst)

            for src in sorted(cycles_by_module.keys()):
                for dst in sorted(cycles_by_module[src]):
                    print(f"  {src} -> {dst}", file=sys.stderr)

            sys.exit(1)
        else:
            print("No circular dependencies found.")
            sys.exit(0)

    # Output results
    if config.json:
        print(json.dumps(results, indent=2))
    elif config.dot:
        print(format_dot(results))
    elif config.sort:
        sorted_modules = topological_sort(results)
        for module in sorted_modules:
            print(module)
    else:
        # Text format
        if len(results) == 1:
            # Single file - just list imports
            print('\n'.join(results[0]['imports']))
        else:
            # Multiple modules - show module names with imports
            for result in results:
                print(f"{result['module']}:")
                for imp in result['imports']:
                    print(f"  {imp}")

    sys.exit(0)

if __name__ == '__main__':
    main(sys.argv)



import json
import os
import pathlib

import pytest

from import_deps import ast_imports
from import_deps import PyModule
from import_deps import ModuleSet
from import_deps.__main__ import main


# list of modules in sample folder used for testing
sample_dir = pathlib.Path(__file__).parent / 'sample-import'
class FOO:
    pkg = sample_dir / 'foo'
    init = pkg / '__init__.py'
    a = pkg / 'foo_a.py'
    b = pkg / 'foo_b.py'
    c = pkg / 'foo_c.py'
    d = pkg / 'foo_d.py'
class SUB:
    pkg = FOO.pkg / 'sub'
    init = pkg / '__init__.py'
    a = pkg / 'sub_a.py'
BAR = sample_dir / 'bar.py'
BAZ = sample_dir / 'baz.py'



def test_ast_imports():
    imports = ast_imports(FOO.a)
    # import bar
    assert (None, 'bar', None, None) == imports[0]
    # from foo import foo_b
    assert ('foo', 'foo_b', None, 0) == imports[1]
    # from foo.foo_c import obj_c
    assert ('foo.foo_c', 'obj_c', None, 0) == imports[2]
    # from .. import sample_d
    assert (None, 'sample_d', None, 2) == imports[3]
    # from ..sample_e import jkl
    assert ('sample_e', 'jkl', None, 2) == imports[4]
    # from sample_f import *
    assert ('sample_f', '*', None, 0) == imports[5]
    # import sample_g.other
    assert (None, 'sample_g.other', None, None) == imports[6]
    # TODO test `impors XXX as YYY`
    assert 7 == len(imports)



class Test_PyModule(object):
    def test_repr(self):
        module = PyModule(SUB.a)
        assert "<PyModule {}>".format(SUB.a) == repr(module)

    def test_is_pkg(self):
        assert True == PyModule.is_pkg(FOO.pkg)
        assert False == PyModule.is_pkg(FOO.init)
        assert False == PyModule.is_pkg(FOO.a)
        assert True == PyModule.is_pkg(SUB.pkg)
        assert False == PyModule.is_pkg(SUB.a)

    def test_fqn(self):
        assert ['bar'] == PyModule(BAR).fqn
        assert ['foo', '__init__'] == PyModule(FOO.init).fqn
        assert ['foo', 'foo_a'] == PyModule(FOO.a).fqn
        assert ['foo', 'sub', 'sub_a'] == PyModule(SUB.a).fqn

    def test_pkg_path(self):
        assert sample_dir == PyModule(BAR).pkg_path()
        assert sample_dir == PyModule(SUB.a).pkg_path()

    def test_relative_path(self):
        cwd = os.getcwd()
        try:
            # do not try to get package beyond given relative path
            os.chdir(FOO.pkg.resolve())
            assert ['foo_a'] == PyModule('foo_a.py').fqn
        finally:
            os.chdir(cwd)

class Test_ModuleSet_Init(object):

    def test_init_with_packge(self):
        modset = ModuleSet([FOO.init, FOO.a])
        assert set(['foo']) == modset.pkgs
        assert 2 == len(modset.by_path)
        assert modset.by_path[FOO.init].fqn == ['foo', '__init__']
        assert modset.by_path[FOO.a].fqn == ['foo', 'foo_a']
        assert 2 == len(modset.by_name)
        assert modset.by_name['foo.__init__'].fqn == ['foo', '__init__']
        assert modset.by_name['foo.foo_a'].fqn == ['foo', 'foo_a']

    def test_init_no_packge(self):
        # if a module of a package is added but no __init__.py
        # its packages is not added to the list of packages
        modset = ModuleSet([FOO.a])
        assert 0 == len(modset.pkgs)
        assert 1 == len(modset.by_path)
        assert modset.by_path[FOO.a].fqn == ['foo', 'foo_a']

    def test_init_subpackge(self):
        modset = ModuleSet([FOO.init, SUB.init, SUB.a])
        assert set(['foo', 'foo.sub']) == modset.pkgs
        assert 3 == len(modset.by_path)
        assert modset.by_path[SUB.a].fqn == ['foo', 'sub', 'sub_a']


class Test_ModuleSet_GetImports(object):

    def test_import_module(self):
        # foo_a  =>  import bar
        modset = ModuleSet([FOO.a, BAR])
        got = modset.get_imports(modset.by_name['foo.foo_a'])
        assert len(got) == 1
        assert BAR in got

    def test_import_not_tracked(self):
        modset = ModuleSet([FOO.a])
        got = modset.get_imports(modset.by_name['foo.foo_a'])
        assert len(got) == 0

    def test_import_pkg(self):
        # bar  =>  import foo
        modset = ModuleSet([FOO.init, BAR])
        got = modset.get_imports(modset.by_name['bar'])
        assert len(got) == 1
        assert FOO.init in got

    def test_from_pkg_import_module(self):
        # foo_a  =>  from foo import foo_b
        modset = ModuleSet([FOO.init, FOO.a, FOO.b])
        got = modset.get_imports(modset.by_name['foo.foo_a'])
        assert len(got) == 1
        assert FOO.b in got

    def test_from_import_object(self):
        # foo_a  =>  from foo.foo_c import obj_c
        modset = ModuleSet([FOO.init, FOO.a, FOO.b, FOO.c])
        got = modset.get_imports(modset.by_name['foo.foo_a'])
        assert len(got) == 2
        assert FOO.b in got # doesnt matter for this test
        assert FOO.c in got

    def test_from_pkg_import_obj(self):
        # baz  =>  from foo import obj_1
        modset = ModuleSet([FOO.init, BAZ])
        got = modset.get_imports(modset.by_name['baz'])
        assert len(got) == 1
        assert FOO.init in got

    def test_import_obj(self):
        # foo_b  =>  import baz.obj_baz
        modset = ModuleSet([FOO.b, BAZ])
        got = modset.get_imports(modset.by_name['foo.foo_b'])
        assert len(got) == 1
        assert BAZ in got

    def test_relative_intra_import_pkg_obj(self):
        # foo_c  =>  from . import foo_i
        modset = ModuleSet([FOO.init, FOO.c])
        got = modset.get_imports(modset.by_name['foo.foo_c'])
        assert len(got) == 1
        assert FOO.init in got

    def test_relative_intra_import_module(self):
        # foo_d  =>  from . import foo_c
        modset = ModuleSet([FOO.init, FOO.c, FOO.d])
        got = modset.get_imports(modset.by_name['foo.foo_d'])
        assert len(got) == 1
        assert FOO.c in got

    def test_relative_parent(self):
        # foo.sub.sub_a  =>  from .. import foo_d
        modset = ModuleSet([FOO.init, FOO.d, SUB.init, SUB.a])
        got = modset.get_imports(modset.by_name['foo.sub.sub_a'])
        assert len(got) == 1
        assert FOO.d in got

    def test_return_module_name(self):
        # foo_a  =>  import bar
        modset = ModuleSet([FOO.a, BAR])
        got = modset.get_imports(modset.by_name['foo.foo_a'],
                                 return_fqn=True)
        name = got.pop()
        assert len(got) == 0
        assert name == 'bar'



    def test_mod_imports(self):
        # foo_a  =>  import bar
        modset = ModuleSet([FOO.init, FOO.a, FOO.b, FOO.c, BAR])
        got = modset.mod_imports('foo.foo_a')
        imports = list(sorted(got))
        assert imports == ['bar', 'foo.foo_b', 'foo.foo_c']


class Test_CLI(object):
    def test_single_file(self, capsys):
        # Test single file analysis
        with pytest.raises(SystemExit) as exc_info:
            main(['import_deps', str(FOO.a)])

        assert exc_info.value.code == 0
        captured = capsys.readouterr()
        lines = captured.out.strip().split('\n')
        assert 'bar' in lines
        assert 'foo.foo_b' in lines
        assert 'foo.foo_c' in lines

    def test_single_file_json(self, capsys):
        # Test single file with JSON output
        with pytest.raises(SystemExit) as exc_info:
            main(['import_deps', str(FOO.a), '--json'])

        assert exc_info.value.code == 0
        captured = capsys.readouterr()
        result = json.loads(captured.out)

        assert len(result) == 1
        assert result[0]['module'] == 'foo.foo_a'
        assert 'bar' in result[0]['imports']
        assert 'foo.foo_b' in result[0]['imports']
        assert 'foo.foo_c' in result[0]['imports']

    def test_directory(self, capsys):
        # Test directory analysis
        with pytest.raises(SystemExit) as exc_info:
            main(['import_deps', str(FOO.pkg)])

        assert exc_info.value.code == 0
        captured = capsys.readouterr()
        output = captured.out

        # Should contain module names and their imports
        assert 'foo.foo_a:' in output
        assert 'foo.foo_d:' in output
        assert 'foo.sub.sub_a:' in output

    def test_directory_json(self, capsys):
        # Test directory with JSON output
        with pytest.raises(SystemExit) as exc_info:
            main(['import_deps', str(FOO.pkg), '--json'])

        assert exc_info.value.code == 0
        captured = capsys.readouterr()
        result = json.loads(captured.out)

        # Should have multiple modules
        assert len(result) > 1

        # Find foo.foo_a module
        foo_a = next((m for m in result if m['module'] == 'foo.foo_a'), None)
        assert foo_a is not None
        assert 'foo.foo_b' in foo_a['imports']
        assert 'foo.foo_c' in foo_a['imports']

        # Find foo.sub.sub_a module
        sub_a = next((m for m in result if m['module'] == 'foo.sub.sub_a'), None)
        assert sub_a is not None
        assert 'foo.foo_d' in sub_a['imports']

    def test_single_file_dot(self, capsys):
        # Test single file with DOT output
        with pytest.raises(SystemExit) as exc_info:
            main(['import_deps', str(FOO.a), '--dot'])

        assert exc_info.value.code == 0
        captured = capsys.readouterr()
        output = captured.out

        # Check DOT format structure
        assert 'digraph imports {' in output
        assert '}' in output
        assert '"foo.foo_a" -> "foo.foo_b";' in output
        assert '"foo.foo_a" -> "foo.foo_c";' in output

    def test_directory_dot(self, capsys):
        # Test directory with DOT output
        with pytest.raises(SystemExit) as exc_info:
            main(['import_deps', str(FOO.pkg), '--dot'])

        assert exc_info.value.code == 0
        captured = capsys.readouterr()
        output = captured.out

        # Check DOT format structure
        assert 'digraph imports {' in output
        assert '}' in output
        assert '"foo.foo_a" -> "foo.foo_b";' in output
        assert '"foo.foo_c" -> "foo.__init__";' in output
        assert '"foo.sub.sub_a" -> "foo.foo_d";' in output

    def test_mutually_exclusive_flags(self, capsys):
        # Test that --json and --dot are mutually exclusive
        with pytest.raises(SystemExit) as exc_info:
            main(['import_deps', str(FOO.a), '--json', '--dot'])

        assert exc_info.value.code == 1
        captured = capsys.readouterr()
        assert 'mutually exclusive' in captured.err

    def test_no_cycles_in_sample(self, capsys):
        # Test that sample data has no circular dependencies
        # So no red edges should appear in DOT output
        with pytest.raises(SystemExit) as exc_info:
            main(['import_deps', str(FOO.pkg), '--dot'])

        assert exc_info.value.code == 0
        captured = capsys.readouterr()
        output = captured.out

        # Normal dependencies should not have color attribute
        assert '"foo.sub.sub_a" -> "foo.foo_d";' in output
        assert '"foo.foo_c" -> "foo.__init__";' in output

        # No cycles in sample data, so no red edges
        assert 'color=red' not in output

    def test_check_no_cycles(self, capsys):
        # Test --check on data without cycles
        with pytest.raises(SystemExit) as exc_info:
            main(['import_deps', str(FOO.pkg), '--check'])

        assert exc_info.value.code == 0
        captured = capsys.readouterr()
        assert 'No circular dependencies found' in captured.out

    def test_sort(self, capsys):
        # Test --sort topological ordering (dependencies before dependents)
        with pytest.raises(SystemExit) as exc_info:
            main(['import_deps', str(FOO.pkg), '--sort'])

        assert exc_info.value.code == 0
        captured = capsys.readouterr()
        modules = captured.out.strip().split('\n')

        # Verify all modules are present
        assert 'foo.__init__' in modules
        assert 'foo.foo_a' in modules
        assert 'foo.foo_b' in modules
        assert 'foo.foo_c' in modules
        assert 'foo.foo_d' in modules
        assert 'foo.sub.__init__' in modules
        assert 'foo.sub.sub_a' in modules

        # Verify topological order: dependencies come before dependents
        # foo.foo_a imports foo.foo_b and foo.foo_c, so they come before foo.foo_a
        assert modules.index('foo.foo_b') < modules.index('foo.foo_a')
        assert modules.index('foo.foo_c') < modules.index('foo.foo_a')

        # foo.foo_c imports foo.__init__, so foo.__init__ comes before foo.foo_c
        assert modules.index('foo.__init__') < modules.index('foo.foo_c')

        # foo.foo_d imports foo.foo_c, so foo.foo_c comes before foo.foo_d
        assert modules.index('foo.foo_c') < modules.index('foo.foo_d')

        # foo.sub.sub_a imports foo.foo_d, so foo.foo_d comes before foo.sub.sub_a
        assert modules.index('foo.foo_d') < modules.index('foo.sub.sub_a')

    def test_sort_mutually_exclusive(self, capsys):
        # Test that --sort and --json are mutually exclusive
        with pytest.raises(SystemExit) as exc_info:
            main(['import_deps', str(FOO.pkg), '--sort', '--json'])

        assert exc_info.value.code == 1
        captured = capsys.readouterr()
        assert 'mutually exclusive' in captured.err

    def test_sort_rank_based_ordering(self):
        # Test that rank-based ordering works correctly
        # Structure: A -> B -> C -> D; B -> E (A imports B, B imports C, etc.)
        # Ranks: D=1, E=1, C=2, B=3, A=4
        # Expected order: D, E, C, B, A (topological: dependencies first, higher rank first when available)
        from import_deps.__main__ import topological_sort

        results = [
            {'module': 'A', 'imports': ['B']},
            {'module': 'B', 'imports': ['C', 'E']},
            {'module': 'C', 'imports': ['D']},
            {'module': 'D', 'imports': []},
            {'module': 'E', 'imports': []},
        ]

        sorted_modules = topological_sort(results)

        # Expected order: D, E, C, B, A
        # D and E are roots (rank 1), D comes before E alphabetically
        # C comes next (rank 2, depends on D)
        # B comes next (rank 3, depends on C and E)
        # A comes last (rank 4, depends on B)
        assert sorted_modules == ['D', 'E', 'C', 'B', 'A']

    def test_sort_with_circular_dependencies(self):
        # Test that circular dependencies are handled gracefully
        # Structure: A -> C -> B -> A (circular); D -> B; E (standalone)
        # (A imports C, C imports B, B imports A, D imports B, E imports nothing)
        from import_deps.__main__ import topological_sort

        results = [
            {'module': 'A', 'imports': ['C']},  # A imports C
            {'module': 'B', 'imports': ['A']},  # B imports A
            {'module': 'C', 'imports': ['B']},  # C imports B (completes cycle)
            {'module': 'D', 'imports': ['B']},  # D imports B (depends on cycle)
            {'module': 'E', 'imports': []},     # E imports nothing (isolated)
        ]

        sorted_modules = topological_sort(results)

        # All modules should be present
        assert len(sorted_modules) == 5
        assert set(sorted_modules) == {'A', 'B', 'C', 'D', 'E'}

        # Cycle nodes (A, B, C) come first (sorted alphabetically), then D, then isolated E
        # D imports B which is in cycle, so D comes after cycle nodes
        # E is isolated, so it comes last
        assert sorted_modules.index('A') < sorted_modules.index('D')
        assert sorted_modules.index('B') < sorted_modules.index('D')
        assert sorted_modules.index('C') < sorted_modules.index('D')
        assert sorted_modules.index('E') == len(sorted_modules) - 1
